
#ifndef __IDE_BUSMASTER_H__
#define __IDE_BUSMASTER_H__

#include "tools.h"

//#define UNI_CPU_OPTIMIZATION

#define QUEUE_STATISTICS
#define MAX_QUEUE_STAT                  8

#define MAX_RETRIES                     3

#define	IO_WD1	        0x1F0		/* Primary Fixed Disk Controller */
#define	IO_WD2	        0x170		/* Secondary Fixed Disk Controller */

#define PCI_ADDRESS_IOMASK              0xfffffff0

#define ATA_BM_OFFSET1			0x08
#define ATA_DMA_ENTRIES			256     /* PAGESIZE/2/sizeof(BM_DMA_ENTRY)*/
#define ATA_DMA_EOT			0x80000000
#define ATA_IOSIZE			0x08
#define ATA_ALTOFFSET			0x206	/* alternate registers offset */
#define ATA_PCCARD_ALTOFFSET		0x0e	/* do for PCCARD devices */
#define ATA_ALTIOSIZE			0x01	/* alternate registers size */
#define ATA_BMIOSIZE			0x20

#define DEV_BSIZE                       512

#define ATAPI_MAGIC_LSB			0x14
#define ATAPI_MAGIC_MSB			0xeb

#define PCI_DEV_HW_SPEC_BM(idhi, idlo) \
    { #idlo, 4, 0x##idlo, #idhi, 4, 0x##idhi}

typedef struct _BUSMASTER_CONTROLLER_INFORMATION {
    PCHAR   VendorId;
    ULONG   VendorIdLength;
    ULONG   nVendorId;
    PCHAR   DeviceId;
    ULONG   DeviceIdLength;
    ULONG   nDeviceId;
    CHAR    VendorIdStr[4];
    CHAR    DeviceIdStr[4];
    ULONG   slotNumber;
    ULONG   busNumber;
    CHAR    channel;
//    CHAR    numOfChannes;
    CHAR    MasterDev;
    BOOLEAN Isr2Enable;
    PDEVICE_OBJECT Isr2DevObj;
    KIRQL   Isr2Irql;
    KAFFINITY  Isr2Affinity;
    ULONG   Isr2Vector;
    PKINTERRUPT Isr2InterruptObject;

}BUSMASTER_CONTROLLER_INFORMATION, *PBUSMASTER_CONTROLLER_INFORMATION;

typedef struct _BUSMASTER_CTX {
    PBUSMASTER_CONTROLLER_INFORMATION* BMListPtr;
    ULONG* BMListLen;
} BUSMASTER_CTX, *PBUSMASTER_CTX;

BUSMASTER_CONTROLLER_INFORMATION const BusMasterAdapters[] = {
    PCI_DEV_HW_SPEC_BM( 1230, 8086 ), //Intel PIIX ATA controller
    PCI_DEV_HW_SPEC_BM( 7010, 8086 ), //Intel PIIX3 ATA controller
    PCI_DEV_HW_SPEC_BM( 7111, 8086 ), //Intel PIIX4 ATA33 controller
    PCI_DEV_HW_SPEC_BM( 7199, 8086 ), // -- // --
    PCI_DEV_HW_SPEC_BM( 84ca, 8086 ), // -- // --
    PCI_DEV_HW_SPEC_BM( 2421, 8086 ), //Intel ICH0 ATA33 controller
    PCI_DEV_HW_SPEC_BM( 2411, 8086 ), //Intel ICH ATA66 controller
    PCI_DEV_HW_SPEC_BM( 7601, 8086 ), // -- // --
    PCI_DEV_HW_SPEC_BM( 244a, 8086 ), //Intel ICH2 ATA100 controller
    PCI_DEV_HW_SPEC_BM( 244b, 8086 ), // -- // --
    PCI_DEV_HW_SPEC_BM( 248a, 8086 ), //Intel ICH3 ATA100 controller
    PCI_DEV_HW_SPEC_BM( 248b, 8086 ), // -- // --

    PCI_DEV_HW_SPEC_BM( 5229, 10b9 ), //AcerLabs Aladdin ATA/ATA33 controller

    PCI_DEV_HW_SPEC_BM( 0571, 1106 ), //VIA ATAxxx controller

    PCI_DEV_HW_SPEC_BM( 5513, 1039 ), //SiS 5591 ATA33 controller

    PCI_DEV_HW_SPEC_BM( 0649, 1095 ), //CMD 649 ATA100 controller
    PCI_DEV_HW_SPEC_BM( 0648, 1095 ), //CMD 648 ATA66 controller
    PCI_DEV_HW_SPEC_BM( 0646, 1095 ), //CMD 646 ATA controller

    PCI_DEV_HW_SPEC_BM( c693, 1080 ), //Cypress 82C693 ATA controller

    PCI_DEV_HW_SPEC_BM( 0102, 1078 ), //Cyrix 5530 ATA33 controller

    PCI_DEV_HW_SPEC_BM( 7409, 1022 ), //AMD 756 ATA66 controller
    PCI_DEV_HW_SPEC_BM( 7411, 1022 ), //AMD 766 ATA100 controller2

    PCI_DEV_HW_SPEC_BM( 0211, 1166 ), //ServerWorks ROSB4 ATA33 controller

    PCI_DEV_HW_SPEC_BM( 4d33, 105a ), //Promise ATA33 controller
    PCI_DEV_HW_SPEC_BM( 4d38, 105a ), //Promise ATA66 controller
    PCI_DEV_HW_SPEC_BM( 0d30, 105a ), //Promise ATA100 controller
    PCI_DEV_HW_SPEC_BM( 4d30, 105a ), // -- // --

    PCI_DEV_HW_SPEC_BM( 4d68, 105a ), //Promise TX2 ATA100 controller
    PCI_DEV_HW_SPEC_BM( 6268, 105a ), //Promise TX2 ATA100 controller
    PCI_DEV_HW_SPEC_BM( 4d69, 105a ), //Promise TX2 ATA133 controller

    PCI_DEV_HW_SPEC_BM( 0004, 1103 )  //HighPoint ATAxxx controller
};

#define NUM_BUSMASTER_ADAPTERS (sizeof(BusMasterAdapters) / sizeof(BUSMASTER_CONTROLLER_INFORMATION))

#define PCI_DEV_CLASS_STORAGE           0x01
#define PCI_DEV_SUBCLASS_IDE            0x01

/* structure for holding DMA address data */
typedef struct BM_DMA_ENTRY {
    ULONG base;
    ULONG count;
} BM_DMA_ENTRY, *PBM_DMA_ENTRY;

typedef struct _IDE_BUSMASTER_REGISTERS {
    UCHAR Command;
    UCHAR DeviceSpecific0;
    UCHAR Status;
    UCHAR DeviceSpecific1;
    ULONG PRD_Table;
} IDE_BUSMASTER_REGISTERS, *PIDE_BUSMASTER_REGISTERS;

#define BM_STATUS_ACTIVE                0x01
#define BM_STATUS_ERR                   0x02
#define BM_STATUS_INTR                  0x04
#define BM_STATUS_MASK                  0x07
#define BM_STATUS_DRIVE_0_DMA           0x20
#define BM_STATUS_DRIVE_1_DMA           0x40
#define BM_STATUS_SIMPLEX_ONLY          0x80

#define BM_COMMAND_START_STOP		0x01
/*#define BM_COMMAND_WRITE		0x08
#define BM_COMMAND_READ		        0x00*/
#define BM_COMMAND_WRITE		0x00
#define BM_COMMAND_READ		        0x08

#define IsBusMaster(pciData) \
    ( ((pciData)->Command & (PCI_ENABLE_BUS_MASTER/* | PCI_ENABLE_IO_SPACE*/)) == \
          (PCI_ENABLE_BUS_MASTER/* | PCI_ENABLE_IO_SPACE*/))

#define PCI_IDE_PROGIF_NATIVE_1         0x01
#define PCI_IDE_PROGIF_NATIVE_2         0x04
#define PCI_IDE_PROGIF_NATIVE_ALL       0x05

#define IsMasterDev(pciData) \
    ( ((pciData)->ProgIf & 0x80) && \
      ((pciData)->ProgIf & PCI_IDE_PROGIF_NATIVE_ALL) != PCI_IDE_PROGIF_NATIVE_ALL )

//#define INT_Q_SIZE 32
#define MIN_REQ_TTL 4

struct _ATA_REQ;

typedef struct _ATA_REQ {
//    ULONG               reqId;          // serial
    struct _ATA_REQ*    next_req;
    struct _ATA_REQ*    prev_req;

    PSCSI_REQUEST_BLOCK Srb;            // Current request on controller.

    PUSHORT             DataBuffer;     // Data buffer pointer.
    ULONG               WordsLeft;      // Data words left.
    LONGLONG            lba;
    ULONG               bcount;

    UCHAR               retry;
    UCHAR               ttl;
//    UCHAR               tag;
    UCHAR               Flags;
    UCHAR               ReqState;

    PSCSI_REQUEST_BLOCK OriginalSrb;    // Mechanism Status Srb Data

    ULONG               dma_base;
    BM_DMA_ENTRY        dma_tab[ATA_DMA_ENTRIES];

} ATA_REQ, *PATA_REQ;

#define REQ_FLAG_FORCE_PIO              0x01
#define REQ_FLAG_DMA_OPERATION          0x02
#define REQ_FLAG_REORDERABLE_CMD        0x04
#define REQ_FLAG_RW_MASK                0x08
#define     REQ_FLAG_READ                   0x08
#define     REQ_FLAG_WRITE                  0x00

// Request states
#define REQ_STATE_NONE                  0x00
#define REQ_STATE_QUEUED                0x10

#define REQ_STATE_PREPARE_TO_TRANSFER   0x20
#define REQ_STATE_READY_TO_TRANSFER     0x30

#define REQ_STATE_EXPECTING_INTR        0x40
#define REQ_STATE_ATAPI_EXPECTING_CMD_INTR     0x41
#define REQ_STATE_ATAPI_EXPECTING_DATA_INTR    0x42

#define REQ_STATE_PROCESSING_INTR       0x50

#define REQ_STATE_DPC_INTR_REQ          0x51
#define REQ_STATE_DPC_RESET_REQ         0x52
#define REQ_STATE_DPC_COMPLETE_REQ      0x53

#define REQ_STATE_TRANSFER_COMPLETE     0x7f

// Command actions:
#define CMD_ACTION_PREPARE              0x01
#define CMD_ACTION_EXEC                 0x02
#define CMD_ACTION_ALL                  (CMD_ACTION_PREPARE | CMD_ACTION_EXEC)

// predefined Reorder costs
#define REORDER_COST_MAX               ((0x1I64 << 60) - 1)
#define REORDER_COST_TTL               (REORDER_COST_MAX - 1)
#define REORDER_COST_INTERSECT         (REORDER_COST_MAX - 2)
#define REORDER_COST_DENIED            (REORDER_COST_MAX - 3)
#define REORDER_COST_RESELECT          (REORDER_COST_MAX/4)

/*typedef struct _ATA_QUEUE {
    struct _ATA_REQ*    head_req; // index
    struct _ATA_REQ*    tail_req; // index
    ULONG               req_count;
    ULONG               dma_base;
    BM_DMA_ENTRY        dma_tab[ATA_DMA_ENTRIES];
} ATA_QUEUE, *PATA_QUEUE;*/

struct _HW_DEVICE_EXTENSION;

// Channel extension
typedef struct _HW_CHANNEL {

    PATA_REQ            first_req;
    PATA_REQ            last_req;
    ULONG               queue_depth;

    BOOLEAN             ExpectingInterrupt;    // Indicates expecting an interrupt
    BOOLEAN             RDP;    // Indicate last tape command was DSC Restrictive.
    // Indicates whether '0x1f0' is the base address. Used
    // in SMART Ioctl calls.
    BOOLEAN             PrimaryAddress;
    // Placeholder for the sub-command value of the last
    // SMART command.
    UCHAR               SmartCommand;
    // Placeholder for status register after a GET_MEDIA_STATUS command
    UCHAR               ReturningMediaStatus;

    MECHANICAL_STATUS_INFORMATION_HEADER MechStatusData;
    SENSE_DATA          MechStatusSense;
    ULONG               MechStatusRetryCount;
    SCSI_REQUEST_BLOCK  InternalSrb;

    ULONG               ChannelCtrlFlags;
    ULONG               ResetInProgress; // flag
    LONG                DisableIntr;
    LONG                CheckIntr;

#define CHECK_INTR_ACTIVE               0x03
#define CHECK_INTR_DETECTED             0x02
#define CHECK_INTR_CHECK                0x01
#define CHECK_INTR_IDLE                 0x00

/*    PHW_TIMER           CurDpc;
    LARGE_INTEGER       ActivationTime;*/

/*    KDPC                Dpc;
    KTIMER              Timer;
    PHW_TIMER           HwScsiTimer;*/
//    KSPIN_LOCK          QueueSpinLock;
//    KIRQL               QueueOldIrql;
    struct _HW_DEVICE_EXTENSION* DeviceExtension;

#ifdef QUEUE_STATISTICS
    LONGLONG QueueStat[MAX_QUEUE_STAT];
    LONGLONG DmaError;
    LONGLONG DmaCrcError;
    LONGLONG PioError;
    LONGLONG ReorderCount;
    LONGLONG IntersectCount;
    LONGLONG TryReorderCount;
    LONGLONG TryReorderHeadCount;
    LONGLONG TryReorderTailCount; /* in-order requests */
#endif //QUEUE_STATISTICS

    //PBM_DMA_ENTRY dma_tab;//[PAGE_SIZE/sizeof(BM_DMA_ENTRY)];	/* DMA transfer table */
//    BM_DMA_ENTRY _dma_tab[PAGE_SIZE/sizeof(BM_DMA_ENTRY)];	/* DMA transfer table */
//    PBM_DMA_ENTRY dma_tab;//[PAGE_SIZE/sizeof(BM_DMA_ENTRY)];	/* DMA transfer table */
//    ULONG dma_base;

} HW_CHANNEL, *PHW_CHANNEL;

#define CTRFLAGS_DMA_ACTIVE             0x01
#define CTRFLAGS_DMA_RO                 0x02
#define CTRFLAGS_DMA_OPERATION          0x04
#define CTRFLAGS_INTR_DISABLED          0x08
#define CTRFLAGS_DPC_REQ                0x10
#define CTRFLAGS_ENABLE_INTR_REQ        0x20

// Logical unit extension
typedef struct _HW_LU_EXTENSION {
    IDENTIFY_DATA2 IdentifyData;
    ULONGLONG      NumOfSectors;
    ULONG          DeviceFlags;    // Flags word for each possible device.
    ULONG          DiscsPresent;   // Indicates number of platters on changer-ish devices.
    BOOLEAN        DWordIO;        // Indicates use of 32-bit PIO
    UCHAR          ReturningMediaStatus;
    UCHAR          TransferMode;
    UCHAR          MaximumBlockXfer;
    ULONG          ErrorCount;     // Count of errors. Used to turn off features.
 //   ATA_QUEUE      cmd_queue;
    LONGLONG       ReadCmdCost;
    LONGLONG       WriteCmdCost;
    LONGLONG       OtherCmdCost;
    //
    PATA_REQ            first_req;
    PATA_REQ            last_req;
    ULONG               queue_depth;

} HW_LU_EXTENSION, *PHW_LU_EXTENSION;

// Device extension
typedef struct _HW_DEVICE_EXTENSION {
    PIDE_REGISTERS_1 BaseIoAddress1[2];    // Base register locations
    PIDE_REGISTERS_2 BaseIoAddress2[2];
    ULONG InterruptLevel;    // Interrupt level
    ULONG InterruptMode;    // Interrupt Mode (Level or Edge)
    // Number of channels being supported by one instantiation
    // of the device extension. Normally (and correctly) one, but
    // with so many broken PCI IDE controllers being sold, we have
    // to support them.
    ULONG NumberChannels;
    HW_LU_EXTENSION lun[4];
    HW_CHANNEL chan[2];
    UCHAR LastInterruptedChannel;
    // Indicates the number of blocks transferred per int. according to the
    // identify data.
    BOOLEAN DriverMustPoll;    // Driver is being used by the crash dump utility or ntldr.
//    BOOLEAN DWordIO;    // Indicates use of 32-bit PIO
    BOOLEAN BusMaster;
    IDENTIFY_DATA FullIdentifyData;    // Identify data for device
    // BusMaster specific data
//    PBM_DMA_ENTRY dma_tab_0;
    KSPIN_LOCK  DpcSpinLock;
    ULONG       ActiveDpcChan;
    ULONG       FirstDpcChan;
    PHW_TIMER   HwScsiTimer1;
    PHW_TIMER   HwScsiTimer2;
    LONGLONG    DpcTime1;
    LONGLONG    DpcTime2;

    PDEVICE_OBJECT Isr2DevObj;

    PIDE_BUSMASTER_REGISTERS  BaseIoAddressBM_0;
    PIDE_BUSMASTER_REGISTERS  BaseIoAddressBM[2];
    ULONG DevID;
    ULONG RevID;
    ULONG Channel;
    ULONG slotNumber;
    ULONG SystemIoBusNumber;
    ULONG DevIndex;
    ULONG HbaCtrlFlags;
    BOOLEAN simplexOnly;
//    BOOLEAN ActiveDpc;
    LONG  ReCheckIntr;
//    KIRQL QueueNewIrql;

} HW_DEVICE_EXTENSION, *PHW_DEVICE_EXTENSION;

typedef struct _ISR2_DEVICE_EXTENSION {
    PHW_DEVICE_EXTENSION HwDeviceExtension;
    ULONG DevIndex;
} ISR2_DEVICE_EXTENSION, *PISR2_DEVICE_EXTENSION;

#define HBAFLAGS_DMA_DISABLED           0x01

//
// values for TransferMode
//
#define         ATA_PIO                 0x00
#define 	ATA_PIO_NRDY            0x01

#define         ATA_PIO0                0x08
#define         ATA_PIO1                0x09
#define         ATA_PIO2                0x0a
#define         ATA_PIO3                0x0b
#define         ATA_PIO4                0x0c
#define         ATA_PIO5                0x0d

#define         ATA_DMA                 0x10
#define         ATA_SDMA                0x10
#define         ATA_SDMA0               0x10
#define         ATA_SDMA1               0x11
#define         ATA_SDMA2               0x12

#define         ATA_WDMA                0x20
#define         ATA_WDMA0               0x20
#define         ATA_WDMA1               0x21
#define         ATA_WDMA2               0x22

#define         ATA_UDMA                0x40
#define         ATA_UDMA0               0x40
#define         ATA_UDMA1               0x41
#define         ATA_UDMA2               0x42
#define         ATA_UDMA4               0x44
#define         ATA_UDMA5               0x45
#define         ATA_UDMA6               0x46

//
//
//
#define         ATA_IDLE                0x0
#define         ATA_IMMEDIATE           0x1
#define         ATA_WAIT_INTR           0x2
#define         ATA_WAIT_READY          0x3
#define         ATA_ACTIVE              0x4
#define         ATA_ACTIVE_ATA          0x5
#define         ATA_ACTIVE_ATAPI        0x6
#define         ATA_REINITING           0x7
#define         ATA_WAIT_BASE_READY     0x8
#define         ATA_WAIT_IDLE           0x9


extern UCHAR         pciBuffer[256];
extern PBUSMASTER_CONTROLLER_INFORMATION BMList;
extern ULONG         BMListLen;

extern VOID
AtapiEnumBusMasterController(
    IN PVOID DriverObject,
    PVOID Argument2
    );

extern ULONG
AtapiFindBusMasterController(
    IN PVOID HwDeviceExtension,
    IN PVOID Context,
    IN PVOID BusInformation,
    IN PCHAR ArgumentString,
    IN OUT PPORT_CONFIGURATION_INFORMATION ConfigInfo,
    OUT PBOOLEAN Again
    );

extern ULONG 
ScsiPortGetBusDataByOffset(
    IN PVOID  HwDeviceExtension,
    IN BUS_DATA_TYPE  BusDataType,
    IN ULONG  BusNumber,
    IN ULONG  SlotNumber,
    IN PVOID  Buffer,
    IN ULONG  Offset,
    IN ULONG  Length
    );

extern ULONG
AtapiFindDev(
    IN PVOID  HwDeviceExtension,
    IN BUS_DATA_TYPE  BusDataType,
    IN ULONG  BusNumber,
    IN ULONG  SlotNumber,
    IN ULONG  dev_id,
    IN ULONG  RevID
    );

extern VOID
AtapiDmaAlloc(
    IN PVOID HwDeviceExtension,
    IN PPORT_CONFIGURATION_INFORMATION ConfigInfo,
    IN ULONG lChannel          // logical channel,
    );

extern BOOLEAN
AtapiDmaSetup(
    IN PVOID HwDeviceExtension,
    IN ULONG DeviceNumber,
    IN ULONG lChannel,          // logical channel,
    IN PSCSI_REQUEST_BLOCK Srb,
    IN PUCHAR data,
    IN ULONG count
    );

extern VOID
AtapiDmaStart(
    IN PVOID HwDeviceExtension,
    IN ULONG DeviceNumber,
    IN ULONG lChannel,          // logical channel,
    IN PSCSI_REQUEST_BLOCK Srb
    );

extern UCHAR
AtapiDmaDone(
    IN PVOID HwDeviceExtension,
    IN ULONG DeviceNumber,
    IN ULONG lChannel,          // logical channel,
    IN PSCSI_REQUEST_BLOCK Srb
    );

extern VOID
AtapiDmaInit(
    IN PVOID HwDeviceExtension,
    IN ULONG DeviceNumber,
    IN ULONG lChannel,          // logical channel,
                               // is always 0 except simplex-only controllers
    IN CHAR apiomode,
    IN CHAR wdmamode,
    IN CHAR udmamode
    );

extern BOOLEAN
AtapiInterrupt2(
    IN PKINTERRUPT Interrupt,
    IN PVOID HwDeviceExtension
    );

extern PDRIVER_OBJECT SavedDriverObject;

/****************** 1 *****************/
#define GetPciConfig1(offs, op) {                                \
    ScsiPortGetBusDataByOffset(HwDeviceExtension,                \
                               PCIConfiguration,                 \
                               SystemIoBusNumber,                \
                               slotNumber,                       \
                               &op,                              \
                               offs,                             \
                               1);                               \
}

#define SetPciConfig1(offs, op) {                                \
    UCHAR _a = op;                                              \
    ScsiPortSetBusDataByOffset(HwDeviceExtension,                \
                               PCIConfiguration,                 \
                               SystemIoBusNumber,                \
                               slotNumber,                       \
                               &_a,                              \
                               offs,                             \
                               1);                               \
}

#define ChangePciConfig1(offs, _op) {                            \
    UCHAR a = 0;                                                \
    GetPciConfig1(offs, a);                                      \
    a = _op;                                                     \
    SetPciConfig1(offs, a);                                      \
}

/****************** 2 *****************/
#define GetPciConfig2(offs, op) {                                \
    ScsiPortGetBusDataByOffset(HwDeviceExtension,                \
                               PCIConfiguration,                 \
                               SystemIoBusNumber,                \
                               slotNumber,                       \
                               &op,                              \
                               offs,                             \
                               2);                               \
}

#define SetPciConfig2(offs, op) {                                \
    USHORT _a = op;                                              \
    ScsiPortSetBusDataByOffset(HwDeviceExtension,                \
                               PCIConfiguration,                 \
                               SystemIoBusNumber,                \
                               slotNumber,                       \
                               &_a,                              \
                               offs,                             \
                               2);                               \
}

#define ChangePciConfig2(offs, _op) {                            \
    USHORT a = 0;                                                \
    GetPciConfig2(offs, a);                                      \
    a = _op;                                                     \
    SetPciConfig2(offs, a);                                      \
}

/****************** 4 *****************/
#define GetPciConfig4(offs, op) {                                \
    ScsiPortGetBusDataByOffset(HwDeviceExtension,                \
                               PCIConfiguration,                 \
                               SystemIoBusNumber,                \
                               slotNumber,                       \
                               &op,                              \
                               offs,                             \
                               4);                               \
}

#define SetPciConfig4(offs, op) {                                \
    ULONG _a = op;                                               \
    ScsiPortSetBusDataByOffset(HwDeviceExtension,                \
                               PCIConfiguration,                 \
                               SystemIoBusNumber,                \
                               slotNumber,                       \
                               &_a,                              \
                               offs,                             \
                               4);                               \
}

#define ChangePciConfig4(offs, _op) {                            \
    ULONG a = 0;                                                 \
    GetPciConfig4(offs, a);                                      \
    a = _op;                                                     \
    SetPciConfig4(offs, a);                                      \
}

#define GetDmaStatus(de, c) \
    (((de)->BaseIoAddressBM[c]) ? ScsiPortReadPortUchar(&((de)->BaseIoAddressBM[c]->Status)) : 0)

#define AtapiPhysToVirtAddr(hwde, srb, phaddr, plen) \
    (ScsiPortConvertPhysicalAddressToUlong/*(ULONG)ScsiPortGetVirtualAddress*/(/*hwde,*/  \
                    ScsiPortGetPhysicalAddress(hwde, srb, phaddr, plen)))

/*#define GET_CHANNEL(Srb)  (Srb->TargetId >> 1)
#define GET_LDEV(Srb)  (Srb->TargetId)
#define GET_LDEV2(P, T, L)  (T)*/

#define GET_CHANNEL(Srb)  (Srb->PathId)
#define GET_LDEV(Srb)  (Srb->TargetId | (Srb->PathId << 1))
#define GET_LDEV2(P, T, L)  (T | (P<<1))


#endif //__IDE_BUSMASTER_H__